#include <math.h>
#include "lis.h"
#include "lis_solver.h"

int lis_bicg(LIS_VECTOR b, LIS_VECTOR x, LIS_VECTOR work[],
		   LIS_SCALAR params[], int options[],
		   LIS_MATRIX A,LIS_MATRIX At, LIS_PRECON M, LIS_SCALAR status[])
{
	LIS_VECTOR r,rtld, z,ztld,p, ptld, q, qtld;
	LIS_SCALAR alpha, beta, rho, rho_old, tmpdot1;
	LIS_SCALAR bnrm2, nrm2;
	LIS_SCALAR tol, rhotol;
	int iter,maxiter,output,n,np;
	double times,ptimes;

	n       = A->n;
	np      = A->np + options[LIS_OPTIONS_PADDING]; 
	maxiter = options[LIS_OPTIONS_MAXITER];
	output  = options[LIS_OPTIONS_OUTPUT];
	tol     = params[LIS_PARAMS_RESID];

	r       = work[0];
	rtld    = work[1];
	z       = work[2];
	ztld    = work[3];
	p       = work[4];
	ptld    = work[5];
	q       = work[2];
	qtld    = work[3];
	rho_old = (LIS_SCALAR)1.0;

	ptimes  = 0.0;

	// Initial Residual
	if( lis_get_initial_residual(A,NULL,x,b,r,NULL,tol,&bnrm2,status) )
	{
		return LIS_SUCCESS;
	}
	lis_vector_copy(r,rtld);
	lis_vector_set_all(0, p);
	lis_vector_set_all(0, ptld);

	for( iter=1; iter<=maxiter; iter++ )
	{
		/* z    = M^-1 * r */
		/* ztld = M^-T * rtld */
		times = lis_wtime();
		lis_psolve(M, r, z);
		lis_psolvet(M, rtld, ztld);
		ptimes += lis_wtime()-times;

		/* rho = <z,rtld> */
		lis_vector_dot(z,rtld,&rho);

		/* test breakdown */
		if( rho==0.0 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_BREAKDOWN;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			return LIS_BREAKDOWN;
		}

		/* beta = (rho / rho_old) */
		beta = rho / rho_old;

		/* p    = z    + beta*p    */
		/* ptld = ztld + beta*ptld */
		
		/* q    = A   * p    */
		/* qtld = A^T * ptld */
		lis_vector_xpay(z,beta,p);
		lis_matvec(A,p,q);

		lis_vector_xpay(ztld,beta,ptld);
		lis_matvect(At,ptld,qtld);

		
		/* tmpdot1 = <ptld,q> */
		lis_vector_dot(ptld,q,&tmpdot1);

		/* test breakdown */
		if( tmpdot1==0.0 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_BREAKDOWN;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			return LIS_BREAKDOWN;
		}
		
		/* alpha = rho / tmpdot1 */
		alpha = rho / tmpdot1;
		
		/* x = x + alpha*p */
		lis_vector_axpy(alpha,p,x);
		
		/* r    = r    - alpha*q    */
		lis_vector_axpy(-alpha,q,r);
		
		/* convergence check */
		lis_vector_nrm2(r,&nrm2);
		nrm2 = nrm2 * bnrm2;

		if( output )
		{
			if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
			if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
		}

		if( tol >= nrm2 )
		{
			status[LIS_STATUS_RET]    = (LIS_SCALAR)LIS_SUCCESS;
			status[LIS_STATUS_ITER]   = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID]  = nrm2;
			status[LIS_STATUS_PTIMES] = ptimes;
			return LIS_SUCCESS;
		}
		
		/* rtld = rtld - alpha*qtld */
		lis_vector_axpy(-alpha,qtld,rtld);

		rho_old = rho;
	}

	status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_MAXITER;
	status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
	status[LIS_STATUS_RESID] = nrm2;
	return LIS_MAXITER;
}
